#include "ofApp.h"
#include "ofAppBaseWindow.h"

//--------------------------------------------------------------
void ofApp::setup(){

        ofBackground(0,0,0);

        camWidth 		= 640;	// try to grab at this size.
        camHeight 		= 480;

        //Counter variable for colorpicking
        cCounter = 0;

        font.load("arial.ttf", 6);

        vidGrabber.setVerbose(true);
       // vidGrabber.setup(camWidth,camHeight);

        //we can now get back a list of devices.
        vector<ofVideoDevice> devices = vidGrabber.listDevices();

        for(int i = 0; i < devices.size(); i++){
            if(devices[i].bAvailable){
                ofLogNotice() << devices[i].id << ": " << devices[i].deviceName;
            }else{
                ofLogNotice() << devices[i].id << ": " << devices[i].deviceName << " - unavailable ";
            }
        }

        //choose camera and and intitialise it
        vidGrabber.setDeviceID(0); // guvcview -o -d /dev/video1
        vidGrabber.setDesiredFrameRate(50);
        vidGrabber.initGrabber(camWidth, camHeight);

        outPixels.allocate(camWidth, camHeight, OF_PIXELS_RGB);
        videoTexture.allocate(outPixels);
        ofSetVerticalSync(true);

        //Arduino setups
        arduino.connect("/dev/ttyACM0", 57600);

        // listen for EInitialized notification. this indicates that
        // the arduino is ready to receive commands and it is safe to
        // call setupArduino()
        ofAddListener(arduino.EInitialized, this, &ofApp::setupArduino);
        bSetupArduino	= false;	// flag so we setup arduino when its ready
}

//------------------------------------------------------------
void ofApp::setupArduino(const int & version) {

    // remove listener because we don't need it anymore
    ofRemoveListener(arduino.EInitialized, this, &ofApp::setupArduino);

    // it is now safe to send commands to the Arduino
    bSetupArduino = true;

    // print firmware name and version to the console
    ofLogNotice() << arduino.getFirmwareName();
    ofLogNotice() << "firmata v" << arduino.getMajorFirmwareVersion() << "." << arduino.getMinorFirmwareVersion();

    // Note: pins A0 - A5 can be used as digital input and output.
    // Refer to them as pins 14 - 19 if using StandardFirmata from Arduino 1.0.
    // If using Arduino 0022 or older, then use 16 - 21.
    // Firmata pin numbering changed in version 2.3 (which is included in Arduino 1.0)

    // set pins D3, D9 and D10 as PWM
    arduino.sendDigitalPinMode(3, ARD_PWM);
    arduino.sendDigitalPinMode(9, ARD_PWM);
    arduino.sendDigitalPinMode(10, ARD_PWM);

    //Simulating initialising of copter with remote (gas-stick up and down)
    for (int i = 0; i < 255; i++) {
          arduino.sendPwm(10, i);
            delay(0.00001);
        }
        for (int i = 255; i > 0; i--) {
          arduino.sendPwm(10, i);
            delay(0.00001);
        }

    //Creates a log file at path '"youropenFrameworks"/apps/myApps/ProjectBOB/bin/data/log.log.txt'
    //All following ofLog() commands append to this file
    ofLogToFile("log/log.txt", true);
    ofLogNotice() << "Setup completed. Program starting" << endl;
}

//--------------------------------------------------------------
//update() reprats itself after each run
void ofApp::update(){

    vidGrabber.update();
    //check if new frame varies from the one before
    if(vidGrabber.isFrameNew()){
        updateScreen();

        if(!bPause && bSetupArduino && !bTrim && !bOutOfPerimeter)
            moveToTarget();
        else if(bOutOfPerimeter && pwmZ > 1)
            pwmZ /= 1.03;
        if(bPause)
            pwmZ = 0;
        arduino.sendPwm(10, pwmZ, true);

//________________________________________
// Trimming was not required in te last version, since moveToMouse was renewed. The new movement function can balance itself
// quite well. It does work anyways and can be used if desired.
    //    if(onTheMove == FALSE && bSetupArduino == TRUE && bTrim == TRUE)
    //        trim(target[0], target[1]);
//________________________________________

        videoTexture.loadData(outPixels);
        arduino.update();
    } else if(bOutOfPerimeter && pwmZ > 1)
        pwmZ -= 2;
}

//--------------------------------------------------------------
void ofApp::draw(){
    vidGrabber.draw(5,5);

    // 2. Image (custom color)
    videoTexture.draw(10 + camWidth, 5, camWidth, camHeight);

    //super green nightvision elite military-grade drone camera interface
    ofSetColor(10, 200, 10);

    string posMsg;
    posMsg += "CopterPos: " + ofToString(center[0]) + " | " + ofToString(center[1]) + "\n";
    posMsg += "TargetPos:  " + ofToString(target[0]) + " | " + ofToString(target[1]) + "\n";
    font.drawString(posMsg, 660, 470);

    string moveMsg;
    moveMsg += "pwmZ: " + ofToString(pwmZ) + "\n";
    moveMsg += "pwmX:    " + ofToString(pwmX) + "\n";
    moveMsg += "pwmY:    " + ofToString(pwmY) + "\n";
    moveMsg += "Trim:   " + ofToString(trimX) + " | " + ofToString(trimY) + "\n";
    moveMsg += "TargetDistance: " + ofToString(targetDist[0]) + " | " + ofToString(targetDist[1]);
    font.drawString(moveMsg, 1100, 450);

    //Draw red target rectangle
    ofSetColor(200, 10, 10);
    ofDrawCircle(target[0] + camWidth, target[1], 2);
    ofDrawLine(target[0] + camWidth-5, target[1]-4, target[0] + camWidth-2, target[1] - 5);
    ofDrawLine(target[0] + camWidth+2, target[1]-4, target[0] + camWidth+5, target[1] - 5);
    ofDrawLine(target[0] + camWidth-5, target[1]-4, target[0] + camWidth-5, target[1]+5);
    ofDrawLine(target[0] + camWidth-5, target[1]+6, target[0] + camWidth-2, target[1]+5);
    ofDrawLine(target[0] + camWidth+2, target[1]+6, target[0] + camWidth+5, target[1]+5);
    ofDrawLine(target[0] + camWidth+6, target[1]-4, target[0] + camWidth+5, target[1]+5);

    //reset color for future drawings
    ofSetColor(100, 100, 100);

    ofDrawLine(center[0] + camWidth, center[1], center[0] + camWidth - 127 + pwmX, center[1] + 127 - pwmY);
}


//--------------------------------------------------------------
void ofApp::keyPressed(int key){

    /* Various keys for controlling the program: (!) american keyboard layout (!)
     *  'P': Pause - Unpause (Stops the copter immediately)
     *  'O': Set pwmZ to 120
     *  '+': increase height / increase pwmZ
     *  '-': decrease height / decrease pwmZ
     *  's', 'w', 'e', 'd': changing p1 and p2 (experimental)
     *  "Arrow keys": Manual software trimming (up-down for pwmY, left-right for pwmX)
     */

    switch(key){
    case 'p': case 'P':
        if(bPause){
            bPause = FALSE;
            pwmZ = pwmV;
        } else {
            bPause = TRUE;
            pwmV = pwmZ;
            pwmZ = 0;
        }
        break;
    case 'o':
        pwmZ = 122;
    case '+': case '=':
        optimalDistance += 2;
        cout << "optDist: " << optimalDistance << endl;
//           pwmZ += 5;
//            if(pwmZ > 255)
//            pwmZ = 255;
        break;
    case '-': case '_':
        optimalDistance -= 2;
        cout << "optDist: " << optimalDistance << endl;
//        pwmZ -= 2;
//        if(pwmZ <= 0)
//            pwmZ = 0;
        break;
    case 's': case 'S':
        p2 /= 1.1;
        ofLogNotice() << "p2: " << p2;
        cout << "p2: " << p2 << endl;
        break;
    case 'w': case 'W':
        p2 *= 1.1;
        ofLogNotice() << "p2 : " << p2;
        cout << "p2: " << p2 << endl;
        break;
    case 'd': case 'D':
        p1 /= 1.1;
        ofLogNotice() << "p1: " << p1;
        cout << "p1: " << p1 << endl;
        break;
    case 'e': case 'E':
        p1 *= 1.1;
        ofLogNotice() << "p1: " << p1;
        cout << "p1: " << p1 << endl;
        break;
    case OF_KEY_LEFT:
        trimX -= 2;
        if(trimX < 0)
            trimX = 0;
        break;
    case OF_KEY_RIGHT:
        trimX += 2;
        if(trimX > 255)
            trimX = 255;
        break;
    case OF_KEY_UP:
        trimY += 2;
        if(trimY > 255)
            trimY = 255;
        break;
    case OF_KEY_DOWN:
        trimY -= 2;
        if(trimY < 0)
            trimY = 0;
        break;
    default:
        break;
    }
}

//--------------------------------------------------------------
void ofApp::keyReleased(int key){

}

//--------------------------------------------------------------
void ofApp::mouseMoved(int x, int y ){

}

//--------------------------------------------------------------
void ofApp::mouseDragged(int x, int y, int button){

}

//--------------------------------------------------------------
void ofApp::mousePressed(int x, int y, int button){

    //event for RMB pressed: pick colors
    if(button == 2) {
    if(cCounter == 2) cCounter = 0;
        if(cCounter == 0) {

            ofPixels &pixels = vidGrabber.getPixels();

            unsigned int red = pixels[((x + y * pixels.getWidth()) * 3)      + 0];        // get red color from mouseposition
            unsigned int green = pixels[((x + y * pixels.getWidth()) * 3)    + 1];       //  get the green color
            unsigned int blue = pixels[((x + y * pixels.getWidth()) * 3)     + 2];       //  get blue

            cout << "Selected color at pos: " << "x: " << x << " y: " << y << endl;



            gaugeColor[0] = red;
            gaugeColor[1] = green;
            gaugeColor[2] = blue;

            cout << "Red: " << red << "\nGreen: " << green << "\nBlue: " << blue << endl;

            cCounter++;
        } else {

            ofPixels &pixels = vidGrabber.getPixels();

            unsigned int red = pixels[((x + y * pixels.getWidth()) * 3)      + 0];        // get red color from mouseposition
            unsigned int green = pixels[((x + y * pixels.getWidth()) * 3)    + 1];       //  get the green color
            unsigned int blue = pixels[((x + y * pixels.getWidth()) * 3)     + 2];       //  get blue

            cout << "Selected color at pos: x2: " << x << " y2: " << y << endl;



            gaugeColor2[0] = red;
            gaugeColor2[1] = green;
            gaugeColor2[2] = blue;


            cout << "Red2: " << red << "\ngreen2: " << green << "\nblue2: " << blue << endl;

            cCounter++;
        }
    }

    //event for LMB pressed: save position
    if(button == 0){
        if(x>655) x -= 640; // enabling clicking on right screen
        target[0] = x;
        target[1] = y;

//______________________________________________________________
        // If copter is on target position - trim
//        if(x-center[0] < 10 &&  x-center[0] > -10 && y-center[1] < 10 && y-center[1] > -10 && onTheMove == FALSE){
//            bTrim = TRUE;
//            pwmZ = 127;
//            int trimDistX = 0;
//            int trimDistY = 0;
//        }
//______________________________________________________________


        if(bPause && !bTrim){
            bPause = FALSE;
            if(pwmZ == 0) //TODO: check if needed
                pwmZ = 120;// ---
            oldPosX = center[0];
            oldPosY = center[1];
            oldTime = ofGetElapsedTimeMillis();
        }
    }

}

//--------------------------------------------------------------
void ofApp::mouseReleased(int x, int y, int button){

}

//--------------------------------------------------------------
void ofApp::mouseEntered(int x, int y){

}

//--------------------------------------------------------------
void ofApp::mouseExited(int x, int y){

}

//--------------------------------------------------------------
void ofApp::windowResized(int w, int h){

}

//--------------------------------------------------------------
void ofApp::gotMessage(ofMessage msg){

}

//--------------------------------------------------------------
void ofApp::dragEvent(ofDragInfo dragInfo){

}


//<><><><><<><><><><><><><><><><><><><><><><><><><><><><>
//  Non-openFramework-functions from here on.          //
//  These can can be called, but wont activate         //
//  by occured events.                                 //
//<><><><>><><><><><>><<><><><><><<>><><><<><><><><><<><>
//
//
//--------------------------------------------------------------
// calculate various locations
void ofApp::getCalc(int redX, int redY, int greenX, int greenY, int redCounter, int greenCounter){
    greenX /= greenCounter;                                              //get center of all green spots
    greenY /= greenCounter;
    redX /= redCounter;                                                  //get center of all red spots
    redY /= redCounter;

    coAngle = atan2 (greenY - redY, greenX - redX);                      //get angle of new coordinate system S'

    center[0] = (greenX + redX)/2 + 5;                                      //Center between signals, X, Y coordinates in old system S
    center[1] = (greenY + redY)/2 + 5;

    greenX -= center[0];                                                //choose Center as origin in old coordinate system S
    greenY -= center[1];
    redX -= center[0];
    redY -= center[1];

    greenCenter[0] = greenX * cos(coAngle) + greenY * sin(coAngle);       //Center for green signal, X, Y coordinates in S'
    greenCenter[1] = -greenX * sin(coAngle) + greenY * cos(coAngle);

    redCenter[0] = redX * cos(coAngle) + redY * sin(coAngle);             //Center for red signal, X, Y coordinates in S'
    redCenter[1] = -redX * sin(coAngle) + redY * cos(coAngle);

    distance[0] = greenX-redX;
    distance[1] = greenY-redY;                          //distance between red and green signal

    currentDistance = sqrt(pow(distance[0], 2) + pow(distance[1], 2));      //absolute distance between signal lights
//    height = 310 - 5775 / currentDistance;

//    cout << "currentDist: " << currentDistance << endl;
//    cout << "Height: " << height << endl;

    targetDist[0] = (target[0]-center[0]) * cos(coAngle) + (target[1]-center[1]) * sin(coAngle);    //shift and turn targetposition
    targetDist[1] = -(target[0]-center[0]) * sin(coAngle) + (target[1]-center[1]) * cos(coAngle);
}

//--------------------------------------------------------------

void ofApp::updateScreen(){

    unsigned int greenX = 0;
    unsigned int greenY = 0;
    unsigned int redX = 0;
    unsigned int redY = 0;
    int redcounter = 0;
    int greencounter = 0;

    ofPixels &pixels = vidGrabber.getPixels();

    int pixelIndex=0;
    // the pixels are aranged in a pixel array, enumerating line by line
    // each pixel consists of 3 following 'unsigned char' values
    // the 3 numbers ranging from 0-255 display the brightness of 'red' 'green' 'blue' (rgb)

    for(int y=0;y<camHeight;y++){
        for(int x=0;x<camWidth;x++){

            // performant way to get color values of each pixel
            unsigned char red = pixels[pixelIndex];pixelIndex++; // get red value from array and move one position further
            unsigned char green = pixels[pixelIndex];pixelIndex++; //get green value from array and move one position further
            unsigned char blue = pixels[pixelIndex];pixelIndex++; //get blue value from array and move one position further

            int indexForPos=(x+y*pixels.getWidth())*3;

            // Check if current Color equals GaugeColor1, if so, color it
            if(abs(red - gaugeColor[0]) + abs(green - gaugeColor[1]) + abs(blue - gaugeColor[2]) <= 50) {
                outPixels[indexForPos+0]=red;
                outPixels[indexForPos+1]=green;
                outPixels[indexForPos+2]=blue;

                //Find and save all coordinates of red pixels - else find green
                if (gaugeColor[0]>gaugeColor[1]){
                    redX += x;
                    redY += y;
                    redcounter ++;
                } else if (gaugeColor[1]>gaugeColor[0]){
                    greenX += x ;
                    greenY += y;
                    greencounter++;
                }

            // Check if current Color equals GaugerColor2, if so, color it
            } else if(abs(red - gaugeColor2[0]) + abs(green - gaugeColor2[1]) + abs(blue - gaugeColor2[2]) <= 50) {
                        outPixels[indexForPos+0]=red;
                        outPixels[indexForPos+1]=green;
                        outPixels[indexForPos+2]=blue;

                        if (gaugeColor2[0]>gaugeColor2[1]){
                            redX += x;
                            redY += y;
                            redcounter ++;

                        } else if (gaugeColor2[1]>gaugeColor2[0]){
                            greenX += x ;
                            greenY += y;
                            greencounter++;
                        }
            // If not Color(1/2), paint it black
            } else {
                outPixels[indexForPos+0]=10;
                outPixels[indexForPos+1]=10;
                outPixels[indexForPos+2]=10;
            }
        }
    }
    //Find center points of red signal, green signal and in between
    if(redcounter != 0 && greencounter != 0){
        getCalc(redX, redY, greenX, greenY, redcounter, greencounter);
    }

    if(greencounter < 10 || redcounter < 10){
        bOutOfPerimeter = TRUE;
    } else bOutOfPerimeter = FALSE;
}

//------------------------------------------------------------------------------
void ofApp::trim(int positionX, int positionY){
    bool bRemain = 1;

    arduino.sendPwm(10, pwmZ);
    arduino.sendPwm(3, trimX);
    arduino.sendPwm(9, trimY);

    cout << "triming..." << endl;

    if(center[0] - positionX > trimDistX + 3 || center[0] - positionX < trimDistX - 3){
        bRemain = 0;
        trimX -= center[0] - positionX - trimDistX;
        if(trimX > 255){
//            trimX += center[0] - positionX - trimDistX;
            trimX = 255;
            cout << "Reached maximum trim!" << endl;
        }
        if(trimX < 1){
            trimX = 1;
            cout << "Reached minimum trim!" << endl;
        }
        arduino.sendPwm(3, trimX);
    }

    if(center[1] - positionY > trimDistY + 3 || center[1] - positionY < trimDistY + 3){
        bRemain = 0;
        trimY += center[1] - positionY - trimDistY;
        if(trimY > 255){
//            trimY -= center[1] - positionY - trimDistY;
            trimY = 255;
            cout << "Reached maximum trim!" << endl;
        }
        if(trimY < 1){
            trimY = 1;
            cout << "Reached minimum trim!" << endl;
        }
        arduino.sendPwm(9, trimY);
    } else if(bRemain){
        bTrim = FALSE;
        for(int i = pwmZ; i>0; --i){
                pwmZ = i;
        }
    }

    trimDistX = center[0] - positionX;
    trimDistY = center[1] - positionY;
}

void ofApp::moveToTarget(){

    // PWM signals depend on the distance to the target and the current speed of the copter
    pwmX = -p2 * (-p1 * targetDist[0] - (oldPosX - center[0]) / ((ofGetElapsedTimef() - oldTime)));
    pwmY = p2 * (-p1 * targetDist[1] - (oldPosY - center[1]) / ((ofGetElapsedTimef() - oldTime)));

    pwmZ = p2V * ((optimalDistance - currentDistance) - ((currentDistance - oldDistance) / p1V) / (ofGetElapsedTimef() - oldTime));
    cout << "currentDist: " << currentDistance << endl;

    //manual trim values are added to PWM signal. Ensuring PWMs range exactly between 0-255
    pwmX += trimX; if(pwmX > 255) pwmX = 255; if(pwmX < 1) pwmX = 1;
    pwmY += trimY; if(pwmY > 255) pwmY = 255; if(pwmY < 1) pwmY = 1;
    pwmZ += optimalDistance * 4; if(pwmZ > 255) pwmZ = 255; if(pwmZ < 1) pwmZ = 0;

    //Log PWMs, distance to target and velocity into a file as a table
    ofLogNotice() << "\t" << pwmX << "\t " << pwmY << "\t" /* << "pwmZ: " << pwmZ  << "\t" */
                  << (short)targetDist[0] << "\t" << (short)targetDist[1] << "\t"
                  << (short)((oldPosX - center[0]) / (ofGetElapsedTimef() - oldTime))<< "\t"
                  << (short)((oldPosY - center[1]) / (ofGetElapsedTimef() - oldTime));


    //send PWM signals to Arduino
    arduino.sendPwm(10, pwmZ, true);
    arduino.sendPwm(3, pwmX, false);
    arduino.sendPwm(9, pwmY, false);


    oldTime = ofGetElapsedTimef();
    oldPosX = center[0];
    oldPosY = center[1];

}
void ofApp::delay(double sleep){
    oldTime = ofGetElapsedTimef();
    while(ofGetElapsedTimef() - oldTime < sleep){
    }
}
